import greenfoot.*;

/**
 * An actor which ca be in a ScrollWorld.
 * It has 3 locations:
 * <Li>
 * A location where it really is. That means, the
 * location the space the camera is moving over.
 * This is the most important one. Access it by
 * the {@link getGlobalX} and the {@link getGlobalY}
 * methods.
 * </Li><Li>
 * A location seen from the camera. This means when
 * it's at {@link 0,0}, it is at the center of what
 * you see. Access it by the {@link getXFromCamera}
 * and the {@link getYFromCamera} methods.
 * </Li><Li>
 * A location on screen. This is what you would do in
 * your usual scenarios, so {@link 0,0} is the top left
 * corner. With this library, it is the least important
 * one. Access it by the usual {@link getX} and {@link
 * getY} methods.
 * </Li>
 * 
 * @author Sven van Nigtevecht
 * @version 2.0
 */
public abstract class ScrollActor extends Actor
{
    private int camX, camY;
    private int globalX, globalY;
    
    /**
     * Returns the x coordinate where you're really
     * standing.
     */
    public int getGlobalX()
    {
        return globalX;
    }
    
    /**
     * Returns the y coordinate where you're really
     * standing.
     */
    public int getGlobalY()
    {
        return globalY;
    }
    
    /**
     * Returns your x coördinate seen from the camera.
     */
    public int getXFromCamera()
    {
        return camX;
    }
    
    /**
     * Returns your y coördinate seen from the camera.
     */
    public int getYFromCamera()
    {
        return camY;
    }
    
    /**
     * Sets your location seen from the world
     * (regular location).
     * That means that a negative location is off
     * screen, and bigger than the world's size is
     * also off screen.
     */
    public void setLocation(int x, int y)
    {
        super.setLocation(x,y);
        int halfWorldWidth = getWorld().getBackground().getWidth() /2;
        int halfWorldHeight = getWorld().getBackground().getHeight() /2;
        camX = x -halfWorldWidth;
        camY = y -halfWorldHeight;
        globalX = x +(getWorld().getCameraX() -halfWorldWidth);
        globalY = y +(getWorld().getCameraY() -halfWorldHeight);
    }
    
    /**
     * Sets your location seen from the camera.
     */
    public void setLocationFromCamera(int x, int y)
    {
        setLocation(x +getWorld().getCameraX(), y +getWorld().getCameraY());
    }
    
    /**
     * Sets your location in the big space (the
     * space where the camera is moving over).
     */
    public void setGlobalLocation(int x, int y)
    {
        GreenfootImage back = getWorld().getBackground();
        int subX = getWorld().getCameraX() -back.getWidth() /2;
        int subY = getWorld().getCameraY() -back.getHeight() /2;
        setLocation(x -subX, y -subY);
    }
    
    /**
     * Returns the ScrollWorld you're in.
     * If you're not in a scroll world (or a subclass of it),
     * and you're a subclass of ScrollActor, and you use
     * this method, this will throw an exception.
     */
    public ScrollWorld getWorld()
    {
        return (ScrollWorld) super.getWorld();
    }
}