import greenfoot.*;
import java.util.ArrayList;

/**
 * A world which has the possibility to scroll
 * over a really big area. That big area is called
 * the big space, It scrolls over it using a camera.
 * The camera's location starts as half width, half height of
 * the world (top left corner of the big space). It can't get
 * outside of the big space.<p>
 * 
 * The background is always scrolling, you don't have
 * to do anything for it. Just make a background like you
 * always do in your scenarios, and this scenario will
 * scroll over it.<p>
 * 
 * It supports adding ScrollActors in 2 ways:
 * <Li>
 * Adding it in the regular way, which means that it can
 * go off screen and come back on screen.
 * </Li><Li>
 * Adding a camera follower, which means that it will
 * move just as much as the camera does. So if you add it
 * to the center of the screen, it will always be there 
 * (except for when you change it yourself of course :) ).
 * </Li>
 * IMPORTANT NOTE: if your Actor isn't a subclass of ScrollActor, it
 * will looklike it is a camera follower.
 * 
 * @author Sven van Nigtevecht
 * @version 2.0
 */
public abstract class ScrollWorld extends World
{
    private final int width, height;
    private ArrayList<ScrollActor> objects;
    private ArrayList<ScrollActor> camFollowers;
    private int fullWidth, fullHeight;
    
    private int camX, camY, camDir;
    
    private GreenfootImage bigBackground;
    private int scrollPosX, scrollPosY;
    
    /**
     * Create a new ScrollWorld.
     * @param width The width of the scroll world in cells.
     * @param height The height of the scroll world in cells.
     * @param cellSize The size of the cells (in pixels).
     * @param fullWidth The total width of the world.
     * That means, objects can't move further than this limit.
     * @param fullHeihgt The total height of the world.
     * That means, objects can't move further than this limit.
     * @throws IllegalArgumentException If one of the arguments
     * is smaller or equal to 0.
     */
    public ScrollWorld(int width, int height, int cellSize, int fullWidth, int fullHeight)
    {
        super(width, height, cellSize, false);
        this.width = getBackground().getWidth();
        this.height = getBackground().getHeight();
        this.fullWidth = fullWidth;
        this.fullHeight = fullHeight;
        if (fullWidth <= 0)
            throw new IllegalArgumentException("The total width of the world ("+fullWidth
            +") can't be smaller or equal to 0.");
        if (fullHeight <= 0)
            throw new IllegalArgumentException("The total height of the world ("+fullHeight
            +") can't be smaller or equal to 0.");
        
        objects = new ArrayList<ScrollActor>();
        camFollowers = new ArrayList<ScrollActor>();
        
        scrollPosX = 0;
        scrollPosY = 0;
        camX = getWidth() /2;
        camY = getHeight() /2;
        camDir = 0;
        
        initializeScrollingBackground();
    }
    
    /** EXTRA METHODS: */
    
    /**
     * Initializes some stuff to make the scrolling background
     * possible. If you ever set your background after calling
     * 'super(...)' in your constructor, you must call this method.
     */
    protected void initializeScrollingBackground()
    {
        GreenfootImage back = getBackground();
        
        bigBackground = new GreenfootImage(width *2, height *2);
        
        bigBackground.drawImage(back, 0,0);
        bigBackground.drawImage(back, width,0);
        bigBackground.drawImage(back, 0,height);
        bigBackground.drawImage(back, width,height);
        
        back.drawImage(bigBackground, scrollPosX, scrollPosY);
    }
    
    /** ADDING + REMOVING OBJECTS: */
    
    /**
     * Adds an object which will follow the camera.
     * The location is seen from the camera, not from the
     * big space.
     * @param cameraFollower The object that will be added to the world
     * as a camera follower.
     * @param x The x coördinate seen from the camera where the object
     * will be added.
     * @param y The y coördinate seen from the camera where the object
     * will be added.
     */
    public void addCameraFollower(ScrollActor cameraFollower, int x, int y)
    {
        super.addObject(cameraFollower, width /2 +x, height /2 +y);
        camFollowers.add(cameraFollower);
    }
    
    /**
     * Adds an object to the big space. Note that the
     * location is also in the big space. If the location is outside
     * it, this will add it to the edge.
     * @param object The object that will be added to the world.
     * @param x The x coördinate in the world where the object
     * will be added.
     * @param y The y coördinate in the world where the object
     * will be added.
     */
    public void addObject(ScrollActor object, int x, int y)
    {
        if (x >= fullWidth)
            x = fullWidth -1;
        else if (x < 0)
            x = 0;
        if (y >= fullHeight)
            y = fullHeight -1;
        else if (y < 0)
            y = 0;
        super.addObject(object, x -(camX -width /2), y -(camY -height /2));
        objects.add(object);
    }
    
    /**
     * Removes an object from the world.
     * @param object The object that will be removed
     * from the world. This can either be a camera follower,
     * or just a regular object.
     */
    public void removeObject(ScrollActor object)
    {
        super.removeObject(object);
        objects.remove(object);
        camFollowers.remove(object);
    }
    
    /** RETURN VALUES: */
    
    /**
     * Returns the camera's x coördinate in big space.
     */
    public int getCameraX()
    {
        return camX;
    }
    
    /**
     * Returns the camera's y coördinate in big space.
     */
    public int getCameraY()
    {
        return camY;
    }
    
    /**
     * Returns the width of the big space.
     */
    public int getFullWidth()
    {
        return fullWidth;
    }
    
    /**
     * Returns the height of the big space.
     */
    public int getFullHeight()
    {
        return fullHeight;
    }
    
    /** CAMERA MOVEMENT + ROTATION: */
    
    /**
     * Moves the camera to a particular location.
     * Note that this is a location in the big space.
     * @param x The new x coördinate of the camera.
     * @param y The new y coördinate of the camera.
     */
    public void moveCameraTo(int x, int y)
    {
        if (camX == x && camY == y) return;
        if (x > fullWidth -getBackground().getWidth() /2)
            x = fullWidth -getBackground().getWidth() /2;
        else if (x < getBackground().getWidth() /2)
            x = getBackground().getWidth() /2;
        if (y > fullHeight -getBackground().getHeight() /2)
            y = fullHeight -getBackground().getHeight() /2;
        else if (y < getBackground().getHeight() /2)
            y = getBackground().getHeight() /2;
        int dx = x -camX;
        int dy = y -camY;
        camX = x;
        camY = y;
        for (ScrollActor a : objects) {
            a.setLocation(a.getX() -dx, a.getY() -dy);
        }
        for (ScrollActor a : camFollowers) {
            a.setLocation(a.getX(), a.getY());
        }
        moveBackgroundRight(dx);
        moveBackgroundUp(dy);
    }
    
    /**
     * Sets the direction the camera is facing.
     * It doesn't change anything you see, but it makes
     * it possible to use the {@link moveCamera} method.
     * @param degrees The new rotation in degrees.
     */
    public void setCameraDirection(int degrees)
    {
        if (camDir == degrees) return;
        camDir = degrees;
        if (camDir >= 360) {
            while (camDir >= 360)
                camDir -= 360;
        } else if (camDir < 0) {
            while (camDir < 0)
                camDir += 360;
        }
    }
    
    /**
     * Turns the camera a bit.
     * It doesn't change anything you see, but it makes
     * it possible to use the {@link moveCamera} method.
     * @param amount The number of degrees the camera will
     * turn clockwise. If this is negative, it will turn
     * counter-clockwise.
     */
    public void turnCamera(int amount)
    {
        setCameraDirection(camDir +amount);
    }
    
    /**
     * Moves the camera forward to the direction
     * it's facing.
     * @param amount The number of cells the camera will move.
     * When this is negative, the camera will move forward.
     */
    public void moveCameraForward(int amount)
    {
        if (amount == 0) return;
        double radians = Math.toRadians(camDir);
        double dx = Math.cos(radians) *amount;
        double dy = Math.sin(radians) *amount;
        moveCameraTo((int)(camX +dx +0.5), (int)(camY +dy +0.5));
    }
    
    /** MOVING BACKGROUND: */
    
    /**
     * All the honor for this goes to Busch2207 from
     * greenfoot.org
     */
    private void moveBackgroundUp(int amount)
    {
        if (amount == 0) return;
        int height = getHeight();
        scrollPosY -= amount;
        while (scrollPosY < 0) {
            scrollPosY += height;
        }
        scrollPosY %= height;
        getBackground().drawImage(bigBackground,
                                  scrollPosX -getWidth(),
                                  scrollPosY -height);
    }
    
    /**
     * All the honor for this goes to Busch2207 from
     * greenfoot.org
     */
    private void moveBackgroundRight(int amount)
    {
        if (amount == 0) return;
        int width = getWidth();
        scrollPosX -= amount;
        while (scrollPosX < 0) {
            scrollPosX += width;
        }
        scrollPosX %= width;
        getBackground().drawImage(bigBackground,
                                  scrollPosX -width,
                                  scrollPosY -getHeight());
    }
}
